"""
Test for the expansion of packages.
"""
from logging import LogRecord
from typing import List, Mapping, Optional

import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]
from maus.edifact import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableDataProvider
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.expression_resolver import expand_packages
from ahbicht.expressions.package_expansion import JsonFilePackageResolver, PackageResolver
from ahbicht.mapping_results import PackageKeyConditionExpressionMapping, Repeatability, parse_repeatability
from unittests.defaults import DefaultPackageResolver, return_empty_dummy_evaluatable_data


class TestPackageResolver:
    """
    Test for the expansions of packages
    """

    @pytest.fixture
    def inject_package_resolver(self, request: SubRequest):
        result_dict: Mapping[str, Optional[str]] = request.param
        resolver = DefaultPackageResolver(result_dict)
        inject.clear_and_configure(
            lambda binder: binder.bind(  # type:ignore[arg-type]
                TokenLogicProvider, SingletonTokenLogicProvider([resolver])
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        "inject_package_resolver",
        [{"123P": "[1] U ([2] O [3])"}],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "unexpanded_expression, expected_expanded_expression",
        [
            pytest.param("[123P]", "[1] U ([2] O [3])"),
            pytest.param("[123P7..8]", "[1] U ([2] O [3])"),
            pytest.param("[123P10..11]", "[1] U ([2] O [3])"),
            pytest.param("[123P0..5]", "[1] U ([2] O [3])"),
            pytest.param("[17] U [123P]", "[17] U ([1] U ([2] O [3]))"),
        ],
    )
    async def test_correct_injection(
        self, inject_package_resolver, unexpanded_expression: str, expected_expanded_expression: str
    ):
        unexpanded_tree = parse_condition_expression_to_tree(unexpanded_expression)
        actual_tree = await expand_packages(parsed_tree=unexpanded_tree)
        assert actual_tree is not None
        expected_expanded_tree = parse_condition_expression_to_tree(expected_expanded_expression)
        assert actual_tree == expected_expanded_tree

    @pytest.mark.parametrize(
        "inject_package_resolver",
        [{"123P": "[1] U ([2] O [3])"}],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "unexpanded_expression, error_message",
        [
            pytest.param("[123P8..7]", "0≤n≤m is not fulfilled for n=8, m=7"),
        ],
    )
    async def test_invalid_package_repeatability(
        self, inject_package_resolver, unexpanded_expression: str, error_message: str
    ):
        unexpanded_tree = parse_condition_expression_to_tree(unexpanded_expression)
        with pytest.raises(ValueError) as invalid_repeatability_error:
            _ = await expand_packages(parsed_tree=unexpanded_tree)
        assert error_message in str(invalid_repeatability_error)

    @pytest.mark.datafiles("./unittests/provider_test_files/example_package_mapping_dict.json")
    @pytest.mark.datafiles("./unittests/provider_test_files/example_package_mapping_list.json")
    @pytest.mark.parametrize(
        "filename",
        [pytest.param("example_package_mapping_dict.json"), pytest.param("example_package_mapping_list.json")],
    )
    async def test_file_based_package_resolver(self, caplog, filename, datafiles):
        """Tests if package resolver provider is instantiated correctly."""
        path_to_hint_json = datafiles / filename
        package_resolver: PackageResolver = JsonFilePackageResolver(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            file_path=path_to_hint_json,
        )
        assert package_resolver.edifact_format == EdifactFormat.UTILMD
        assert package_resolver.edifact_format_version == EdifactFormatVersion.FV2104
        assert await package_resolver.get_condition_expression("123P") == PackageKeyConditionExpressionMapping(
            edifact_format=EdifactFormat.UTILMD, package_key="123P", package_expression="[1] U [2] O [3]"
        )
        assert await package_resolver.get_condition_expression("456P") == PackageKeyConditionExpressionMapping(
            edifact_format=EdifactFormat.UTILMD, package_key="456P", package_expression="[4] U [5] X [6]"
        )
        log_entries: List[LogRecord] = caplog.records
        assert log_entries[0].message == "Instantiated JsonFilePackageResolver"
        assert log_entries[1].message.startswith("Resolved expression")

    def test_how_to_access_repeatability_token(self):
        """
        This test demonstrates how to access the repeatability token from the parsed tree
        We extract those tokens also in the PackageExpansionTransformer, but we don't do anything with them there yet.
        """
        unexpanded_tree = parse_condition_expression_to_tree("[1P2..3]")
        assert unexpanded_tree is not None
        repeatability_tokens = [token for token in unexpanded_tree.children if token.type == "REPEATABILITY"]
        assert parse_repeatability(repeatability_tokens[0]) == Repeatability(min_occurrences=2, max_occurrences=3)



"""
AHBicht is a lark based parser and evaluation framework for conditions that occur in Anwendungshandbüchern (AHB).
"""
import sys
from enum import Enum

if sys.version_info.major == 3 and sys.version_info.minor >= 11:
    from enum import StrEnum

    # We have to use the builtin / std lib enum.StrEnum in Python >= 3.11, because the behaviour of (str,Enum) changed:
    # class Foo(str, Enum):
    #     MEMBER = "MEMBER"
    # f"{a_str_enum_member}" results in "MEMBER" for Python < v3.11 but "Foo.MEMBER" in Python >= v3.11
    # Using the old (str, Enum) in Python 3.11 causes errors e.g. in the ExpressionBuilders.
else:

    class StrEnum(str, Enum):  # type:ignore[no-redef]
        """
        An enum class of which each member has a string representation.
        This is a workaround for Python <v3.11 because enum.StrEnum was introduced in Python 3.11.
        """

        # We'll live with this  class for Python <v3.11. The unit test for python 3.9-3.11 ensure that this works.



"""
Tests that the code can handle RC/FC evaluators that have both async and sync methods.
"""

import inject
import pytest  # type:ignore[import]

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider, EvaluationContext
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from unittests.defaults import (
    default_test_format,
    default_test_version,
    empty_default_hints_provider,
    return_empty_dummy_evaluatable_data,
)


class MixedSyncAsyncRcEvaluator(RcEvaluator):
    """An RC evaluator that has both sync and async methods"""

    edifact_format = default_test_format
    edifact_format_version = default_test_version

    def _get_default_context(self):
        return None  # type:ignore[return-value]

    def evaluate_1(self, evaluatable_data, context):
        assert isinstance(evaluatable_data, EvaluatableData)
        if context is not None:
            assert isinstance(context, EvaluationContext)
        return ConditionFulfilledValue.FULFILLED

    async def evaluate_2(self, evaluatable_data, context):
        assert isinstance(evaluatable_data, EvaluatableData)
        if context is not None:
            assert isinstance(context, EvaluationContext)
        return ConditionFulfilledValue.UNFULFILLED


class MixedSyncAsyncFcEvaluator(FcEvaluator):
    """An FC Evaluator that has both sync and async methods"""

    edifact_format = default_test_format
    edifact_format_version = default_test_version

    def evaluate_901(self, _):
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)

    async def evaluate_902(self, _):
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)


class TestMixedSyncAsyncEvaluation:
    @pytest.mark.parametrize(
        "expression, expected_rc_fulfilled, expected_fc_fulfilled",
        [
            pytest.param("Muss ([1][901] X [2][902])", True, True),
            pytest.param("Muss ([1][902] X [2][901])", True, True),
            pytest.param("Muss ([2][902] X [2][901])", False, True),
        ],
    )
    async def test_mixed_async_non_async(
        self, expression: str, expected_rc_fulfilled: bool, expected_fc_fulfilled: bool
    ):
        fc_evaluator = MixedSyncAsyncFcEvaluator()
        rc_evaluator = MixedSyncAsyncRcEvaluator()
        inject.clear_and_configure(
            lambda binder: binder.bind(  # type:ignore[arg-type]
                TokenLogicProvider,
                SingletonTokenLogicProvider([rc_evaluator, fc_evaluator, empty_default_hints_provider]),
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        tree = await parse_expression_including_unresolved_subexpressions(expression)
        evaluation_result = await evaluate_ahb_expression_tree(tree)
        assert (
            evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            is expected_rc_fulfilled
        )
        assert (
            evaluation_result.format_constraint_evaluation_result.format_constraints_fulfilled is expected_fc_fulfilled
        )
        # When mocker.spy ing on the evaluate_... the inspection inside the rc/fc evaluator fails
        # https://stackoverflow.com/questions/18869141/using-python-mock-to-spy-on-calls-to-an-existing-object



"""
This module provides the functions to validate segment groups, segments and data elements.
"""

import asyncio
from typing import Awaitable, Dict, List, Optional

from maus.models.anwendungshandbuch import DeepAnwendungshandbuch
from maus.models.edifact_components import (
    DataElement,
    DataElementDataType,
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    SegmentLevel,
)

from ahbicht.content_evaluation import fc_evaluators
from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResult,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions import InvalidExpressionError
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.enums import ModalMark, PrefixOperator, RequirementIndicator
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.validation import validation_logger
from ahbicht.validation.validation_results import (
    DataElementValidationResult,
    SegmentLevelValidationResult,
    ValidationResultInContext,
)
from ahbicht.validation.validation_values import RequirementValidationValue


async def validate_deep_anwendungshandbuch(
    deep_ahb: DeepAnwendungshandbuch, soll_is_required: bool = True
) -> List[ValidationResultInContext]:
    """
    Validates a deep Anwendungshandbuch aka 'MAUS' as provided from the package maus.
    :param deep_ahb: the deep Anwendungshandbuch that should be validated
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: List of ValidationResultInContext of the deep Anwendungshandbuch
    """

    tasks: List[Awaitable[List[ValidationResultInContext]]] = []

    for segment_group in deep_ahb.lines:
        tasks.append(
            validate_segment_group(
                segment_group=segment_group,
                soll_is_required=soll_is_required,
            )
        )

    validation_results_of_segment_groups: List[List[ValidationResultInContext]] = await asyncio.gather(*tasks)
    deep_ahb_validation_result: List[ValidationResultInContext] = []
    for sublist in validation_results_of_segment_groups:
        for item in sublist:
            deep_ahb_validation_result.append(item)
    validation_logger.debug("Validated %i segment groups", len(deep_ahb_validation_result))
    return deep_ahb_validation_result


async def validate_segment_level(
    segment_level: SegmentLevel,
    soll_is_required: bool = True,
) -> List[ValidationResultInContext]:
    """
    Validates the root segment group or segment and its children by handing them
    over to specialized functions for each kind.
    :param segment_level: the segment group or segment that should be validated
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: List of ValidationResultInContext of the segment or segment group including its children
    """
    if isinstance(segment_level, SegmentGroup):
        return await validate_segment_group(segment_group=segment_level, soll_is_required=soll_is_required)

    if isinstance(segment_level, Segment):
        return await validate_segment(segment=segment_level, soll_is_required=soll_is_required)

    raise NotImplementedError("Unexpected type for segment_level")


async def validate_segment_group(
    segment_group: SegmentGroup,
    parent_segment_group_requirement: Optional[RequirementValidationValue] = None,
    soll_is_required: bool = True,
) -> List[ValidationResultInContext]:
    """
    Validates a segment group and the segment groups and segments it contains.
    :param segment_group: the segment_group that should be validated
    :param parent_segment_group_requirement: the requirement of the segment_group's segment_group, e.g. IS_REQUIRED
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: List of ValidationResultInContext of the segment group and the segment groups and segments it contains.
    """

    # validation of this segment group
    if parent_segment_group_requirement is RequirementValidationValue.IS_FORBIDDEN:
        segment_group_validation = SegmentLevelValidationResult(
            requirement_validation=RequirementValidationValue.IS_FORBIDDEN
        )
    else:
        segment_group_validation = await get_segment_level_requirement_validation_value(
            segment_group, parent_segment_group_requirement, soll_is_required
        )

    validation_results_in_context = [
        ValidationResultInContext(discriminator=segment_group.discriminator, validation_result=segment_group_validation)
    ]

    if segment_group_validation.requirement_validation is not RequirementValidationValue.IS_FORBIDDEN:
        tasks: List[Awaitable[List[ValidationResultInContext]]] = []

        # validation of child_segment_group s
        if segment_group.segment_groups:
            for child_segment_group in segment_group.segment_groups:
                tasks.append(
                    validate_segment_group(
                        child_segment_group,
                        segment_group_validation.requirement_validation,
                        soll_is_required,
                    )
                )

        # validation of child segments
        if segment_group.segments:
            for segment in segment_group.segments:
                tasks.append(
                    validate_segment(
                        segment,
                        segment_group_validation.requirement_validation,
                        soll_is_required,
                    )
                )

        validation_results_of_children: List[List[ValidationResultInContext]] = await asyncio.gather(*tasks)

        for sublist in validation_results_of_children:
            for item in sublist:
                validation_results_in_context.append(item)

    return validation_results_in_context


async def validate_segment(
    segment: Segment,
    segment_group_requirement: Optional[RequirementValidationValue] = None,
    soll_is_required: bool = True,
) -> List[ValidationResultInContext]:
    """
    Validates a segment and the data elements it contains.
    :param segment: the segment that should be validated
    :param segment_group_requirement: the requirement of the segment's parent segment_group, e.g. IS_REQUIRED
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: List of ValidationResultInContext of the segment and the data elements it contains
    """

    # validation of this segment
    if segment_group_requirement is RequirementValidationValue.IS_FORBIDDEN:
        segment_validation = SegmentLevelValidationResult(
            requirement_validation=RequirementValidationValue.IS_FORBIDDEN
        )
    else:
        segment_validation = await get_segment_level_requirement_validation_value(
            segment, segment_group_requirement, soll_is_required
        )

    validation_results_in_context = [
        ValidationResultInContext(discriminator=segment.discriminator, validation_result=segment_validation)
    ]

    # validate this segments' data elements only if segment is not forbidden
    if segment_validation.requirement_validation is RequirementValidationValue.IS_FORBIDDEN:
        validation_results_in_context_data_elements = []
    else:
        tasks: List[Awaitable[ValidationResultInContext]] = []
        for data_element in segment.data_elements:
            tasks.append(validate_data_element(data_element, segment_validation.requirement_validation))
        validation_results_in_context_data_elements = await asyncio.gather(*tasks)

    return [*validation_results_in_context, *validation_results_in_context_data_elements]


async def get_segment_level_requirement_validation_value(
    segment_level: SegmentLevel,
    parent_segment_group_requirement: Optional[RequirementValidationValue] = None,
    soll_is_required: bool = True,
) -> SegmentLevelValidationResult:
    """
    Parses and evaluates a segment level expression and maps it to a requirement validation value,
    also depending on a parent segment group requirement if present.
    :param segment_level: the segment or segment group that should be validated
    :param parent_segment_group_requirement: the requirement of the segment level's segment group, e.g. IS_REQUIRED
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: Validation Result of the data element
    """
    assert segment_level.ahb_expression is not None
    expression_tree = await parse_expression_including_unresolved_subexpressions(
        segment_level.ahb_expression, resolve_packages=True
    )
    try:
        evaluation_result = await evaluate_ahb_expression_tree(expression_tree)
    except InvalidExpressionError as invalid_expr_error:
        validation_logger.warning("The expression '%s' is invalid. Returning IS_OPTIONAL", segment_level.ahb_expression)
        return SegmentLevelValidationResult(
            hints=invalid_expr_error.error_message, requirement_validation=RequirementValidationValue.IS_OPTIONAL
        )

    requirement_validation_without_hierarchy = map_requirement_validation_values(
        evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled,
        requirement_indicator=evaluation_result.requirement_indicator,
        soll_is_required=soll_is_required,
    )

    requirement_validation = combine_requirements_of_different_levels(
        parent_level_requirement=parent_segment_group_requirement,
        child_level_requirement=requirement_validation_without_hierarchy,
    )

    return SegmentLevelValidationResult(
        requirement_validation=requirement_validation,
        hints=evaluation_result.requirement_constraint_evaluation_result.hints,
    )


async def validate_data_element(
    data_element: DataElement,
    segment_requirement: RequirementValidationValue,
    soll_is_required=True,
) -> ValidationResultInContext:
    """
    Validates data elements by handing them over to specialized functions for freetext or value pool data elements.
    :param data_element: the data element that should be validated
    :param segment_requirement: the requirement of the data element's parent segment, e.g. IS_REQUIRED
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: Validation Result of the Data element
    """
    if isinstance(data_element, DataElementFreeText):
        return await validate_data_element_freetext(data_element, segment_requirement, soll_is_required)
    if isinstance(data_element, DataElementValuePool):
        return await validate_data_element_valuepool(data_element, segment_requirement)
    raise NotImplementedError("Unexpected type for data element")


async def validate_data_element_freetext(
    data_element: DataElementFreeText,
    segment_requirement: Optional[RequirementValidationValue] = None,
    soll_is_required=True,
) -> ValidationResultInContext:
    """
    Validates a freetext data element, e.g. 'Dokumentennummer'.
    :param data_element: the data element that should be validated
    :param segment_requirement: the requirement of the data element's parent segment, e.g. IS_REQUIRED
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, if it should be handled like KANN
    :return: Validation Result of the DataElement
    """
    assert data_element.ahb_expression is not None
    expression_tree = await parse_expression_including_unresolved_subexpressions(
        data_element.ahb_expression, resolve_packages=True
    )
    fc_evaluators.text_to_be_evaluated_by_format_constraint.set(data_element.entered_input)
    try:
        evaluation_result = await evaluate_ahb_expression_tree(expression_tree)
    except InvalidExpressionError as invalid_expr_error:
        validation_logger.warning(
            "The expression '%s' @ '%s' is invalid. Returning IS_OPTIONAL",
            data_element.ahb_expression,
            data_element.discriminator,
        )
        return ValidationResultInContext(
            validation_result=DataElementValidationResult(
                requirement_validation=RequirementValidationValue.IS_OPTIONAL,
                format_validation_fulfilled=True,
                format_error_message=None,
                hints=invalid_expr_error.error_message,
                data_element_data_type=DataElementDataType.TEXT,
            ),
            discriminator=data_element.discriminator,  # type:ignore[arg-type]
        )

    # requirement constraints
    requirement_validation_without_input_without_hierarchy = map_requirement_validation_values(
        evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled,
        evaluation_result.requirement_indicator,
        soll_is_required,
    )
    requirement_validation_without_input = combine_requirements_of_different_levels(
        segment_requirement, requirement_validation_without_input_without_hierarchy
    )

    if data_element.entered_input:
        requirement_validation = RequirementValidationValue(str(requirement_validation_without_input) + "_AND_FILLED")
    else:
        requirement_validation = RequirementValidationValue(str(requirement_validation_without_input) + "_AND_EMPTY")
    result = DataElementValidationResult(
        requirement_validation=requirement_validation,
        format_validation_fulfilled=evaluation_result.format_constraint_evaluation_result.format_constraints_fulfilled,  # pylint: disable=line-too-long
        format_error_message=evaluation_result.format_constraint_evaluation_result.error_message,
        hints=evaluation_result.requirement_constraint_evaluation_result.hints,
        data_element_data_type=DataElementDataType.TEXT,  # set text by default, later check if it might be a date
    )
    if data_element.value_type is not None:
        result.data_element_data_type = data_element.value_type  # this will set the type to DATE, if necessary
    validation_logger.debug(
        "The validation of expression '%s' for the data element with discriminator '%s' resulted in %s",
        data_element.ahb_expression,
        data_element.discriminator,
        str(result),
    )
    return ValidationResultInContext(
        discriminator=data_element.discriminator,  # type:ignore[arg-type]
        validation_result=result,
    )


async def validate_data_element_valuepool(
    data_element: DataElementValuePool,
    segment_requirement: RequirementValidationValue,
) -> ValidationResultInContext:
    """
    Validates a value pool data element which depends on the requirement status of its segment.
    :param data_element: the data element that should be validated
    :param segment_requirement: the requirement of the data element's parent segment, e.g. IS_REQUIRED
    :return: Validation Result of the DataElement
    """
    possible_values: Dict[str, str] = {}

    # Since the conditions behind the qualifiers only say if this qualifier is possible or not,
    # the overall requirement is inherited from the data element's parent segment.
    # Only exception is if no qualifier is possible, see below
    requirement_validation_data_element = segment_requirement

    if segment_requirement is not RequirementValidationValue.IS_FORBIDDEN:
        # it seems like all single value data elements (except freetext) are just labeled "X"
        if len(data_element.value_pool) == 1:
            possible_values[data_element.value_pool[0].qualifier] = data_element.value_pool[0].meaning
            hints = None
        else:  # len(value_pool) >1
            for value_pool_entry in data_element.value_pool:
                expression_tree = await parse_expression_including_unresolved_subexpressions(
                    value_pool_entry.ahb_expression, resolve_packages=True
                )
                try:
                    evaluation_result = await evaluate_ahb_expression_tree(expression_tree)
                except InvalidExpressionError as invalid_expr_error:
                    validation_logger.warning(
                        "The expression '%s' @ '%s' is invalid. Treating value pool entry as optional",
                        value_pool_entry.ahb_expression,
                        data_element.discriminator,
                    )
                    evaluation_result = AhbExpressionEvaluationResult(
                        format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                            format_constraints_fulfilled=True, error_message=None
                        ),
                        requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                            requirement_constraints_fulfilled=True,
                            requirement_is_conditional=True,
                            hints=invalid_expr_error.error_message,
                        ),
                        requirement_indicator=ModalMark.KANN,
                    )
                validation_logger.debug(
                    "The validation of value pool entry %s resulted in %s",
                    str(value_pool_entry),
                    str(evaluation_result),
                )
                if evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled:
                    possible_values[value_pool_entry.qualifier] = value_pool_entry.meaning
                hints = None  # TODO: Get all hints from possible values

    # if no possible values are found, the requirement is set to forbidden
    # pylint: disable = pointless-statement
    # Case: segment_requirement is required, but no possible values are appended
    fc_validation_result: bool = True  # by default Format Constraints are not evaluated for ValuePools
    if not possible_values:
        requirement_validation_data_element is RequirementValidationValue.IS_FORBIDDEN
        hints = None
    else:
        if data_element.entered_input in possible_values:
            requirement_validation_data_element = RequirementValidationValue.IS_REQUIRED_AND_FILLED
        elif data_element.entered_input:
            fc_validation_result = False  # we re-use the fc validation field to mark that the value is unexpected
            requirement_validation_data_element = RequirementValidationValue.IS_REQUIRED_AND_EMPTY
            hints = f"Der Wert '{data_element.entered_input}' ist nicht in: {{{', '.join(possible_values.keys())}}}"
            data_element.entered_input = None  # overwrite the illegal value
        else:
            requirement_validation_data_element = RequirementValidationValue.IS_REQUIRED_AND_EMPTY
    result = DataElementValidationResult(
        requirement_validation=requirement_validation_data_element,
        format_validation_fulfilled=fc_validation_result,
        hints=hints,  # todo: hints might be referenced before assignment
        possible_values=possible_values,
        data_element_data_type=DataElementDataType.VALUE_POOL,
    )
    validation_logger.debug(
        "The validation for the data element with discriminator '%s' resulted in %s",
        data_element.discriminator,
        str(result),
    )
    return ValidationResultInContext(
        discriminator=data_element.discriminator,  # type:ignore[arg-type]
        validation_result=result,
    )


def map_requirement_validation_values(
    requirement_constraints_are_fulfilled: Optional[bool],
    requirement_indicator: RequirementIndicator,
    soll_is_required: bool = True,
) -> RequirementValidationValue:
    """
    Returns requirement validation value according to the requirement indicator
    and whether the requirements constraints are fulfilled.
    :param requirement_constraints_are_fulfilled: true if requirement constraints are fulfilled
    :param requirement_indicator: requirement indicator, e.g MUSS, SOLL, KANN, X, O, U
    :param soll_is_required: true (default) if SOLL should be handled like MUSS, false if it should be handled like KANN
    :return: mapped requirement validation value
    """

    # sets soll according to soll_is_required
    if requirement_indicator is ModalMark.SOLL:
        if soll_is_required:
            requirement_indicator = ModalMark.MUSS
        else:
            requirement_indicator = ModalMark.KANN

    if requirement_constraints_are_fulfilled is False:
        requirement_validation = RequirementValidationValue.IS_FORBIDDEN
    elif requirement_constraints_are_fulfilled is None:
        if requirement_indicator is ModalMark.MUSS or isinstance(requirement_indicator, PrefixOperator):
            # This error is basically the postponed former RequirementEvaluationFailedBecauseOfUnknownNodesError.
            # Raising this error here allows the users to use ConditionFulfilledValue.UNKNOWN in the evaluation module
            # and only care about it again, if they're also using the validation module.
            # https://github.com/Hochfrequenz/ahbicht/issues/275
            raise NotImplementedError(
                # pylint:disable=line-too-long
                "It's unknown if the requirement constraints are met (most likely because at least one node evaluated to UNKNONW. A validation is meaningless in this case."
            )
        if requirement_indicator is ModalMark.KANN or requirement_indicator is ModalMark.SOLL:
            requirement_validation = RequirementValidationValue.IS_OPTIONAL
    else:  # True
        if requirement_indicator is ModalMark.MUSS or isinstance(requirement_indicator, PrefixOperator):
            requirement_validation = RequirementValidationValue.IS_REQUIRED
        elif requirement_indicator is ModalMark.KANN:
            requirement_validation = RequirementValidationValue.IS_OPTIONAL

    return requirement_validation


def combine_requirements_of_different_levels(
    parent_level_requirement: Optional[RequirementValidationValue],
    child_level_requirement: RequirementValidationValue,
) -> RequirementValidationValue:
    """
    Combines two requirement values of a parent and child level.
    Requirement status from `segment group` beats `segment` beats `data element`
    Forbidden on parent level is caught beforehand to make further evaluation unnecessary.

    | parent group | child group | result    |
    | ------------ | ----------- | --------- |
    | required     | required    | required  |
    | required     | forbidden   | forbidden |
    | required     | optional    | optional  |
    | optional     | required    | optional  |
    | optional     | forbidden   | forbidden |
    | optional     | optional    | optional  |

    """
    if parent_level_requirement is None:
        return child_level_requirement

    if parent_level_requirement is RequirementValidationValue.IS_REQUIRED:
        return child_level_requirement
    if parent_level_requirement is RequirementValidationValue.IS_OPTIONAL:
        if child_level_requirement is RequirementValidationValue.IS_REQUIRED:
            return RequirementValidationValue.IS_OPTIONAL  # TODO: Let's discuss this (optional/optionalrequired?)
        return child_level_requirement

    raise ValueError("Unexpected parent_level_requirement value")



""" Tests the PackageResolver, that assumes a ContentEvaluationResult to be present in the evaluatable data"""

import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from ahbicht.mapping_results import PackageKeyConditionExpressionMapping
from unittests.defaults import default_test_format, default_test_version


class TestCerBasedPackageResolver:
    """Test for the evaluation using the ContentEvaluationResult Based Hints Provider"""

    @pytest.mark.parametrize(
        "inject_cer_evaluators",
        [
            pytest.param(
                ContentEvaluationResult(
                    format_constraints={},
                    requirement_constraints={},
                    hints={},
                    packages={
                        "2P": "[2]",
                        "3P": "[3] U [4]",
                    },
                )
            )
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "condition_key, expected_result",
        [
            pytest.param(
                "2P",
                PackageKeyConditionExpressionMapping(
                    package_key="2P", edifact_format=default_test_format, package_expression="[2]"
                ),
            ),
            pytest.param(
                "3P",
                PackageKeyConditionExpressionMapping(
                    package_key="3P", edifact_format=default_test_format, package_expression="[3] U [4]"
                ),
            ),
            pytest.param(
                "4P",
                PackageKeyConditionExpressionMapping(
                    package_expression=None, edifact_format=default_test_format, package_key="4P"
                ),
            ),
        ],
    )
    async def test_evaluation(
        self, condition_key: str, expected_result: PackageKeyConditionExpressionMapping, inject_cer_evaluators
    ):
        token_logic_provider: TokenLogicProvider = inject.instance(TokenLogicProvider)  # type:ignore[assignment]
        package_resolver = token_logic_provider.get_package_resolver(default_test_format, default_test_version)
        actual = await package_resolver.get_condition_expression(condition_key)
        assert actual == expected_result



"""
A module to allow easy distinction between different types of condition nodes (by mapping their integer key)
"""
from ahbicht import StrEnum


class ConditionNodeType(StrEnum):
    """
    Possible types of condition nodes.
    The value is usually determined using the key of the respective condition node.
    For the descriptions of the different types see also the README.
    """

    FORMAT_CONSTRAINT = "FORMAT_CONSTRAINT"
    """
    A Format Constraint (FC) is what edi@energy refers to as "Formatbedingung".
    Other than requirement constraints (RC) format constraints do not refer to a data constellation in the edifact
    message as a whole (or at least parts of it) but to the value used at the place where the format constraint is used.
    """
    REQUIREMENT_CONSTRAINT = "REQUIREMENT_CONSTRAINT"
    """
    A Requirement Constraint (RC) is what edi@energy refers to as "Voraussetzung".
    It describes a data constellation that might (or might not) occur in a edifact message.
    """
    HINT = "HINT"
    """
    A Hint is a node whose text/description starts with "Hinweis". It's a text that cannot be interpreted automatically.
    """
    PACKAGE = "PACKAGE"
    """
    A package is an abbreviation for an expression.
    """
    REPEATABILITY_CONSTRAINT = "REPEATABILITY_CONSTRAINT"
    """
    Repeatability Constraints are what edi@energy refers to as "Wiederholbarkeiten".
    They describe how often you may insert segment (groups) into a message (e.g. min. 1 but max 10 times).
    """


def derive_condition_node_type(condition_key: str) -> ConditionNodeType:
    """
    Returns the corresponding type of condition node for a given condition key
    """
    if condition_key.endswith("P"):
        return ConditionNodeType.PACKAGE
    if 1 <= int(condition_key) <= 499:
        return ConditionNodeType.REQUIREMENT_CONSTRAINT
    if 500 <= int(condition_key) <= 900:
        return ConditionNodeType.HINT
    if 901 <= int(condition_key) <= 999:
        return ConditionNodeType.FORMAT_CONSTRAINT
    if 2000 <= int(condition_key) <= 2499:
        return ConditionNodeType.REPEATABILITY_CONSTRAINT
    raise ValueError("Condition key is not in valid number range.")



import uuid
from itertools import product
from typing import List

import inject
import pytest  # type:ignore[import]
from maus.models.anwendungshandbuch import AhbMetaInformation, DeepAnwendungshandbuch
from maus.models.edifact_components import (
    DataElementDataType,
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.evaluator_factory import create_and_inject_hardcoded_evaluators
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.expressions.enums import ModalMark, PrefixOperator, RequirementIndicator
from ahbicht.validation.validation import (
    combine_requirements_of_different_levels,
    map_requirement_validation_values,
    validate_data_element_freetext,
    validate_data_element_valuepool,
    validate_deep_anwendungshandbuch,
    validate_segment,
    validate_segment_group,
    validate_segment_level,
)
from ahbicht.validation.validation_results import (
    DataElementValidationResult,
    SegmentLevelValidationResult,
    ValidationResultInContext,
)
from ahbicht.validation.validation_values import RequirementValidationValue

# TODO: Add testcases for segment_requirement is None & soll_is_required = False
from unittests.defaults import default_test_format, default_test_version, empty_default_test_data


class TestValidation:
    @pytest.fixture
    def inject_content_evaluation_result(self):
        content_evaluation_result = ContentEvaluationResult(
            hints={"501": "foo"},
            format_constraints={
                "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                "903": EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="Format error 903"),
            },
            requirement_constraints={
                "2": ConditionFulfilledValue.FULFILLED,
                "3": ConditionFulfilledValue.UNFULFILLED,
            },
            id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
        )

        def eval_data_provider():
            return empty_default_test_data

        create_and_inject_hardcoded_evaluators(
            content_evaluation_result=content_evaluation_result,
            edifact_format=default_test_format,
            edifact_format_version=default_test_version,
            evaluatable_data_provider=eval_data_provider,
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        "deep_ahb, expected_validation_result",
        [
            pytest.param(
                DeepAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="12345"),
                    lines=[
                        SegmentGroup(
                            discriminator="root",
                            ahb_expression="X",
                            segments=[
                                Segment(
                                    discriminator="UNH",
                                    ahb_expression="X",
                                    data_elements=[
                                        DataElementFreeText(
                                            discriminator="Nachrichten-Startsegment",
                                            ahb_expression="X",
                                            entered_input=None,
                                            data_element_id="1234",
                                        )
                                    ],
                                )
                            ],
                        ),
                        SegmentGroup(
                            discriminator="SG4",
                            ahb_expression="X",
                            segments=[
                                Segment(
                                    discriminator="FOO",
                                    ahb_expression="X",
                                    data_elements=[
                                        DataElementValuePool(
                                            discriminator="SG4->FOO->0333",
                                            value_pool=[
                                                ValuePoolEntry(
                                                    qualifier="E01",
                                                    meaning="Das andere",
                                                    ahb_expression="X[3]",
                                                ),
                                                ValuePoolEntry(
                                                    qualifier="E02",
                                                    meaning="Das Eine",
                                                    ahb_expression="X[2]",
                                                ),
                                            ],
                                            data_element_id="0333",
                                            entered_input=None,
                                        )
                                    ],
                                )
                            ],
                            segment_groups=[
                                SegmentGroup(
                                    discriminator="SG5",
                                    ahb_expression="X[3]",
                                    segments=[
                                        Segment(
                                            discriminator="BAR",
                                            ahb_expression="X",
                                            data_elements=[
                                                DataElementFreeText(
                                                    discriminator="Die fünfte Gruppe",
                                                    ahb_expression="X",
                                                    entered_input=None,
                                                    data_element_id="1234",
                                                )
                                            ],
                                        )
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
                [
                    ValidationResultInContext(
                        discriminator="root",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="UNH",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED,
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="Nachrichten-Startsegment",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.TEXT,
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG4",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG5",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_FORBIDDEN
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="FOO",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG4->FOO->0333",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            possible_values={"E02": "Das Eine"},
                            data_element_data_type=DataElementDataType.VALUE_POOL,
                        ),
                    ),
                ],
            ),
        ],
    )
    async def test_validate_deep_ahb(
        self,
        deep_ahb,
        expected_validation_result,
        inject_content_evaluation_result,
    ):
        result = await validate_deep_anwendungshandbuch(deep_ahb)
        assert result == expected_validation_result

    @pytest.mark.parametrize(
        "segment_level, expected_validation_result",
        [
            pytest.param(
                SegmentGroup(
                    ahb_expression="Muss[2]",
                    discriminator="SG10",
                    segments=[
                        Segment(
                            ahb_expression="Muss[2]",
                            discriminator="SG10 - Datum",
                            data_elements=[
                                DataElementFreeText(
                                    discriminator="SG10 - Datum - Einzug",
                                    data_element_id="1234",
                                    ahb_expression="M[2]",
                                    entered_input="",
                                    value_type=DataElementDataType.DATETIME,
                                )
                            ],
                        )
                    ],
                    segment_groups=[
                        SegmentGroup(
                            ahb_expression="Muss",
                            discriminator="SG11",
                            segments=[
                                Segment(
                                    ahb_expression="Muss[2]",
                                    discriminator="SG11 - Datum",
                                    data_elements=[
                                        DataElementFreeText(
                                            discriminator="SG11 - Datum - Auszug",
                                            data_element_id="1235",
                                            ahb_expression="M[2]",
                                            entered_input="",
                                            value_type=DataElementDataType.DATETIME,
                                        )
                                    ],
                                )
                            ],
                            segment_groups=None,
                        )
                    ],
                ),
                [
                    ValidationResultInContext(
                        discriminator="SG10",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG11",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG11 - Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG11 - Datum - Auszug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
            ),
            pytest.param(
                Segment(
                    ahb_expression="Muss[2]",
                    discriminator="SG10 - Datum",
                    data_elements=[
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Einzug",
                            data_element_id="1234",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        )
                    ],
                ),
                [
                    ValidationResultInContext(
                        discriminator="SG10 - Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
            ),
        ],
    )
    async def test_validate_root_segment_level(
        self,
        segment_level,
        expected_validation_result,
        inject_content_evaluation_result,
    ):
        result = await validate_segment_level(segment_level)
        assert result == expected_validation_result

    @pytest.mark.parametrize(
        "segment_group, parent_segment_group_requirement, expected_validation_result",
        [
            pytest.param(
                SegmentGroup(
                    ahb_expression="Muss[2]",
                    discriminator="SG10",
                    segments=[
                        Segment(
                            ahb_expression="Muss[2]",
                            discriminator="SG10 - Datum",
                            data_elements=[
                                DataElementFreeText(
                                    discriminator="SG10 - Datum - Einzug",
                                    data_element_id="1234",
                                    ahb_expression="M[2]",
                                    entered_input="",
                                    value_type=DataElementDataType.DATETIME,
                                )
                            ],
                        )
                    ],
                    segment_groups=[],
                ),
                RequirementValidationValue.IS_REQUIRED,
                [
                    ValidationResultInContext(
                        discriminator="SG10",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
            ),
        ],
    )
    async def test_validate_segment_group(
        self,
        segment_group,
        parent_segment_group_requirement,
        expected_validation_result,
        inject_content_evaluation_result,
    ):
        result = await validate_segment_group(segment_group, parent_segment_group_requirement)
        assert result == expected_validation_result

    @pytest.mark.parametrize(
        "segment, segment_group_requirement, expected_validation_result",
        [
            pytest.param(
                Segment(
                    ahb_expression="Muss[2]",
                    discriminator="SG10 Datum",
                    data_elements=[
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Einzug",
                            data_element_id="1234",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        )
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                [
                    ValidationResultInContext(
                        discriminator="SG10 Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
            ),
            pytest.param(
                Segment(
                    ahb_expression="Muss[2]",
                    discriminator="SG10 Datum",
                    data_elements=[
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Einzug",
                            data_element_id="1234",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        ),
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Auszug",
                            data_element_id="1235",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                [
                    ValidationResultInContext(
                        discriminator="SG10 Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Auszug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
            ),
            pytest.param(
                Segment(
                    ahb_expression="Muss[2]O[501]",
                    discriminator="SG10 Datum",
                    data_elements=[
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Einzug",
                            data_element_id="1234",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        ),
                        DataElementFreeText(
                            discriminator="SG10 - Datum - Auszug",
                            data_element_id="1235",
                            ahb_expression="M[2]",
                            entered_input="",
                            value_type=DataElementDataType.DATETIME,
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                [
                    ValidationResultInContext(
                        discriminator="SG10 Datum",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_OPTIONAL,
                            hints="Combining a neutral element 'Hint(condition_key='501', conditions_fulfilled=<ConditionFulfilledValue.NEUTRAL: 'NEUTRAL'>, hint='foo')' with a boolean value RequirementConstraint(condition_key='2', conditions_fulfilled=<ConditionFulfilledValue.FULFILLED: 'FULFILLED'>) in an or_composition is not implemented as it has no useful result. This is probably an error in the AHB itself.",
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Einzug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_OPTIONAL_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="SG10 - Datum - Auszug",
                        validation_result=DataElementValidationResult(
                            requirement_validation=RequirementValidationValue.IS_OPTIONAL_AND_EMPTY,
                            format_validation_fulfilled=True,
                            data_element_data_type=DataElementDataType.DATETIME,
                        ),
                    ),
                ],
                id="invalid ahb expression",
            ),
        ],
    )
    async def test_validate_segment(
        self, segment, segment_group_requirement, expected_validation_result, inject_content_evaluation_result
    ):
        result = await validate_segment(segment, segment_group_requirement)
        assert result == expected_validation_result

    @pytest.mark.parametrize(
        "data_element, segment_requirement, expected_validation",
        [
            pytest.param(
                DataElementFreeText(
                    discriminator="SG1", ahb_expression="X", entered_input="bar", data_element_id="1234"
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                        format_validation_fulfilled=True,
                        data_element_data_type=DataElementDataType.TEXT,
                    ),
                ),
            ),
            pytest.param(
                DataElementFreeText(discriminator="SG1", ahb_expression="X", entered_input="", data_element_id="1234"),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                        format_validation_fulfilled=True,
                        data_element_data_type=DataElementDataType.TEXT,
                    ),
                ),
            ),
            pytest.param(
                DataElementFreeText(
                    discriminator="SG1", ahb_expression="M[903][2]U[501]", entered_input="", data_element_id="1234"
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                        format_validation_fulfilled=False,
                        format_error_message="Format error 903",
                        hints="foo",
                        data_element_data_type=DataElementDataType.TEXT,
                    ),
                ),
            ),
            pytest.param(
                DataElementFreeText(
                    discriminator="SG1", ahb_expression="M[902][3]U[501]", entered_input="bar", data_element_id="1234"
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_FILLED,
                        format_validation_fulfilled=True,
                        data_element_data_type=DataElementDataType.TEXT,
                    ),
                ),
            ),
            pytest.param(
                DataElementFreeText(
                    discriminator="SG1", ahb_expression="M[3]O[501]", entered_input="bar", data_element_id="1234"
                ),
                RequirementValidationValue.IS_OPTIONAL,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_OPTIONAL,
                        format_validation_fulfilled=True,
                        data_element_data_type=DataElementDataType.TEXT,
                        hints="Combining a neutral element 'Hint(condition_key='501', conditions_fulfilled=<ConditionFulfilledValue.NEUTRAL: 'NEUTRAL'>, hint='foo')' with a boolean value RequirementConstraint(condition_key='3', conditions_fulfilled=<ConditionFulfilledValue.UNFULFILLED: 'UNFULFILLED'>) in an or_composition is not implemented as it has no useful result. This is probably an error in the AHB itself.",
                    ),
                ),
            ),
        ],
    )
    async def test_validate_data_element_freetext(
        self, caplog, data_element, segment_requirement, expected_validation, inject_content_evaluation_result
    ):
        result = await validate_data_element_freetext(data_element, segment_requirement)
        assert result == expected_validation
        validation_log_entries = [record for record in caplog.records if record.name == "ahbicht.validation"]
        assert len(validation_log_entries) > 0
        first_log_message = validation_log_entries[0].message
        assert first_log_message.startswith("The validation of expression") or first_log_message.startswith(
            "The expression '"
        )

    @pytest.mark.parametrize(
        "data_element, segment_requirement, expected_validation",
        [
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input=None,
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                        format_validation_fulfilled=True,
                        possible_values={"A1": "Ich bin A1", "A2": "Ich bin A2", "A3": "Ich bin A3"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
                id="required value pool with empty input",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input=None,
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_FORBIDDEN,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_FORBIDDEN,
                        format_validation_fulfilled=True,
                        possible_values={},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
                id="forbidden value pool with empty input",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input=None,
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X[2]",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X[3]",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                        format_validation_fulfilled=True,
                        possible_values={"A1": "Ich bin A1", "A3": "Ich bin A3"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
                id="required value pool and empty input",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input="A2",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X[2]",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X[3]",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                        format_validation_fulfilled=False,
                        possible_values={"A1": "Ich bin A1", "A3": "Ich bin A3"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                        hints="Der Wert 'A2' ist nicht in: {A1, A3}",
                    ),
                ),
                id="required value pool and illegal input",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input=None,
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X[2]",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X[3][501]",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_FORBIDDEN,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_FORBIDDEN,
                        format_validation_fulfilled=True,
                        possible_values={},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input="A1",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X",
                        )
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                        format_validation_fulfilled=True,
                        possible_values={"A1": "Ich bin A1"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
                id="only one expected input (could this be filled automatically?)",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input="A1",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X",
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                        format_validation_fulfilled=True,
                        possible_values={"A1": "Ich bin A1", "A2": "Ich bin A2", "A3": "Ich bin A3"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                    ),
                ),
                id="required value pool and correct input",
            ),
            pytest.param(
                DataElementValuePool(
                    discriminator="SG1",
                    data_element_id="1234",
                    entered_input="A3",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="A1",
                            meaning="Ich bin A1",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A2",
                            meaning="Ich bin A2",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="A3",
                            meaning="Ich bin A3",
                            ahb_expression="X [2] O [501]",  # <-- well formed but invalid
                        ),
                    ],
                ),
                RequirementValidationValue.IS_REQUIRED,
                ValidationResultInContext(
                    discriminator="SG1",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                        format_validation_fulfilled=True,
                        possible_values={"A1": "Ich bin A1", "A2": "Ich bin A2", "A3": "Ich bin A3"},
                        data_element_data_type=DataElementDataType.VALUE_POOL,
                        hints=None,
                    ),
                ),
                id="invalid ahb expression",
            ),
        ],
    )
    async def test_validate_data_element_valuepool(
        self, caplog, data_element, segment_requirement, expected_validation, inject_content_evaluation_result
    ):
        result = await validate_data_element_valuepool(data_element, segment_requirement)
        assert result == expected_validation
        validation_log_entries = [record for record in caplog.records if record.name == "ahbicht.validation"]
        assert len(validation_log_entries) > 0

    @pytest.mark.parametrize(
        "requirement_constraints_are_fulfilled, requirement_indicator, expected_requirement_validation_value",
        [
            pytest.param(
                False,
                ModalMark.MUSS,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
            pytest.param(
                True,
                ModalMark.MUSS,
                RequirementValidationValue.IS_REQUIRED,
            ),
            pytest.param(
                True,
                ModalMark.SOLL,
                RequirementValidationValue.IS_REQUIRED,
            ),
            pytest.param(
                False,
                ModalMark.SOLL,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
            pytest.param(
                True,
                ModalMark.KANN,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                True,
                PrefixOperator.X,
                RequirementValidationValue.IS_REQUIRED,
            ),
            pytest.param(
                False,
                PrefixOperator.X,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
        ],
    )
    def test_map_requirement_validation_values(
        self, requirement_constraints_are_fulfilled, requirement_indicator, expected_requirement_validation_value
    ):
        requirement_validation_value = map_requirement_validation_values(
            requirement_constraints_are_fulfilled, requirement_indicator
        )
        assert requirement_validation_value == expected_requirement_validation_value

    @pytest.mark.parametrize(
        "requirement_constraints_are_fulfilled, requirement_indicator, expected_requirement_validation_value",
        [
            pytest.param(
                False,
                ModalMark.SOLL,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
            pytest.param(
                True,
                ModalMark.SOLL,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                True,
                ModalMark.MUSS,
                RequirementValidationValue.IS_REQUIRED,
            ),
            pytest.param(
                False,
                ModalMark.MUSS,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
            pytest.param(
                None,
                ModalMark.SOLL,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                None,
                ModalMark.KANN,
                RequirementValidationValue.IS_OPTIONAL,
            ),
        ],
    )
    def test_map_requirement_validation_values_soll_not_required(
        self, requirement_constraints_are_fulfilled, requirement_indicator, expected_requirement_validation_value
    ):
        requirement_validation_value = map_requirement_validation_values(
            requirement_constraints_are_fulfilled, requirement_indicator, soll_is_required=False
        )
        assert requirement_validation_value == expected_requirement_validation_value

    def test_map_requirement_validation_values_muss_with_unknown_rcs(self):
        with pytest.raises(NotImplementedError):
            _ = map_requirement_validation_values(None, ModalMark.MUSS, soll_is_required=False)

    def test_map_requirement_validation_values_all_cases_are_covered(self):
        """
        A fuzzing test to make sure all possible input values are mapped
        """
        requirement_indicators: List[RequirementIndicator] = [x for x in ModalMark] + [x for x in PrefixOperator]
        for rcs_fulfilled, requirement_indicator, soll_is_required in product(
            [True, None, False], requirement_indicators, [True, False]
        ):
            try:
                result = map_requirement_validation_values(rcs_fulfilled, requirement_indicator, soll_is_required)
            except NotImplementedError:
                if rcs_fulfilled is None:
                    continue  # this case is ok
                raise
            assert result is not None
            assert isinstance(result, RequirementValidationValue)

    @pytest.mark.parametrize(
        "parent_level_requirement, child_level_requirement, expected_requirement",
        [
            pytest.param(
                RequirementValidationValue.IS_REQUIRED,
                RequirementValidationValue.IS_REQUIRED,
                RequirementValidationValue.IS_REQUIRED,
            ),
            pytest.param(
                RequirementValidationValue.IS_REQUIRED,
                RequirementValidationValue.IS_OPTIONAL,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                RequirementValidationValue.IS_REQUIRED,
                RequirementValidationValue.IS_FORBIDDEN,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
            pytest.param(
                RequirementValidationValue.IS_OPTIONAL,
                RequirementValidationValue.IS_REQUIRED,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                RequirementValidationValue.IS_OPTIONAL,
                RequirementValidationValue.IS_OPTIONAL,
                RequirementValidationValue.IS_OPTIONAL,
            ),
            pytest.param(
                RequirementValidationValue.IS_OPTIONAL,
                RequirementValidationValue.IS_FORBIDDEN,
                RequirementValidationValue.IS_FORBIDDEN,
            ),
        ],
    )
    def test_combine_requirements_of_different_levels(
        self, parent_level_requirement, child_level_requirement, expected_requirement
    ):
        combined_requirement = combine_requirements_of_different_levels(
            parent_level_requirement, child_level_requirement
        )
        assert combined_requirement == expected_requirement

    def test_combine_requirements_of_different_levels_error(self):
        with pytest.raises(ValueError) as excinfo:
            _ = combine_requirements_of_different_levels(
                parent_level_requirement=RequirementValidationValue.IS_FORBIDDEN,
                child_level_requirement=RequirementValidationValue.IS_OPTIONAL,  # does not matter
            )

        assert "Unexpected parent_level_requirement value" in str(excinfo.value)
